@@ -1,5 +1,5 @@
-/*!
- * Copyright 2018-2019 by Contributors
+/**
+ * Copyright 2018-2025, XGBoost Contributors
  */
 #pragma once
 #include <dmlc/io.h>
@@ -14,42 +14,19 @@
 #include "xgboost/feature_map.h"
 #include "xgboost/model.h"
 #include "xgboost/json.h"
-#include "xgboost/parameter.h"
 
 namespace xgboost {
 class Json;
 namespace gbm {
-// Deprecated in 1.0.0. model parameter.  Only staying here for compatible binary model IO.
-struct DeprecatedGBLinearModelParam : public dmlc::Parameter<DeprecatedGBLinearModelParam> {
-  // number of feature dimension
-  uint32_t deprecated_num_feature;
-  // deprecated. use learner_model_param_->num_output_group.
-  int32_t deprecated_num_output_group;
-  // reserved field
-  int32_t reserved[32];
-  // constructor
-  DeprecatedGBLinearModelParam() {
-    static_assert(sizeof(*this) == sizeof(int32_t) * 34,
-                  "Model parameter size can not be changed.");
-    std::memset(this, 0, sizeof(DeprecatedGBLinearModelParam));
-  }
-
-  DMLC_DECLARE_PARAMETER(DeprecatedGBLinearModelParam) {}
-};
-
 // model for linear booster
 class GBLinearModel : public Model {
- private:
-  // Deprecated in 1.0.0
-  DeprecatedGBLinearModelParam param_;
-
  public:
-  int32_t num_boosted_rounds;
+  std::int32_t num_boosted_rounds{0};
   LearnerModelParam const* learner_model_param;
 
  public:
-  explicit GBLinearModel(LearnerModelParam const* learner_model_param) :
-      num_boosted_rounds{0}, learner_model_param {learner_model_param} {}
+  explicit GBLinearModel(LearnerModelParam const *learner_model_param)
+      : learner_model_param{learner_model_param} {}
   void Configure(Args const &) { }
 
   // weight for each of feature, bias is the last one
@@ -68,17 +45,6 @@ class GBLinearModel : public Model {
   void SaveModel(Json *p_out) const override;
   void LoadModel(Json const &in) override;
 
-  // save the model to file
-  void Save(dmlc::Stream *fo) const {
-    fo->Write(&param_, sizeof(param_));
-    fo->Write(weight);
-  }
-  // load model from file
-  void Load(dmlc::Stream *fi) {
-    CHECK_EQ(fi->Read(&param_, sizeof(param_)), sizeof(param_));
-    fi->Read(&weight);
-  }
-
   // model bias
   inline bst_float *Bias() {
     return &weight[learner_model_param->num_feature *
@@ -122,7 +88,7 @@ class GBLinearModel : public Model {
         }
       }
       fo << std::endl << "    ]" << std::endl << "  }";
-    } else {
+    } else if (format == "text") {
       fo << "bias:\n";
       for (int gid = 0; gid < ngroup; ++gid) {
         fo << this->Bias()[gid] << std::endl;
@@ -133,6 +99,8 @@ class GBLinearModel : public Model {
           fo << (*this)[i][gid] << std::endl;
         }
       }
+    } else {
+      LOG(FATAL) << "Dump format `" << format << "` is not supported by the gblinear model.";
     }
     std::vector<std::string> v;
     v.push_back(fo.str());
